(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsModel GpsModelLemmas GpsLogic.
Require Import GpsGlobal GpsVisible GpsRelyGhost GpsGuarPrep.

Local Open Scope string_scope.

Set Implicit Arguments.

(******************************************************************************)
(** * 3.3.7. Instrumented execution and adequacy *)
(******************************************************************************)

Definition erase (x : actid * exp * resource * (nat -> RProp)) := 
   let '(a, e, r, QQ) := x in (a, e).

Theorem gsafe_pres :
  forall n mc L T mc',
    gsafe (n + 1) (Macts mc) (Mamap mc) (Msb mc) (Mrf mc) (Mmo mc) L T ->
    map erase T = Mtmap mc ->
    mc_step mc mc' ->
    exists L' T',
      gsafe n (Macts mc') (Mamap mc') (Msb mc') (Mrf mc') (Mmo mc') L' T' /\
      map erase T' = Mtmap mc' /\
      exists rem, map snd T' = map snd T ++ rem.
Proof.
  destruct 3, mc as [? acts], mc' as [? ? amap sb rf mo]; ins; desf.

Case "normal step".
  eapply map_eq_app_inv in THR; desf.
  destruct l2' as [|[[? r] QQ] l2']; ins; desf.
  red in H; desf.
  exploit SAFE; eauto using in_or_app, in_eq.
  rewrite <- plus_n_Sm, <- plus_n_O; ins; desf.
  clear END FORK.
  rename a' into b.
  assert (ND' : NoDup (b :: acts)) by vauto; clear ND.

  assert (sb a b) by (rewrite SB; auto).
  assert (a <> b) by (intro; red in CONS; desc; edestruct ACYC; vauto).
  exploit (SBS a).
    by red in CONS; desc; red in FIN; desc; edestruct CLOsb as [? _]; eauto; ins; desf. 
  rewrite filter_app, map_app, rsum_app; simpl; rewrite eqxx; simpl; rewrite res_plusAC.
  intro LAB.
  (* 1. Prepare *)
  assert (PI  := Consistent_pres_in CONS).

  exploit prepare_step; eauto.
  clear AEQ SB RF MO.
  clear LAB LR CC CO V; rename L into Lold; intros (L & ?); desf.
  clear Mamap Msb Mrf Mmo CONS0.
  generalize (fun a' IN NEQ => eq_trans (LEQ a' NEQ) (SBS a' IN)); 
    clear Lold SBS LEQ; intro SBS.

  assert (SBS': forall a', In a' acts ->
              L (TsbS a') =
              rsum (map (fun aerP : nat * _ * _  * _ => snd (fst aerP))
                        (filter (fun aerP => fst (fst (fst aerP)) == a')
                                (l1' ++ l2')))).
  by clear - LAB SBS; intros; destruct (eqP a' a); desf; 
     [rewrite LAB | rewrite SBS; ins]; rewrite !filter_app, !map_app, !rsum_app; 
     ins; desf; desf.
  clear LAB SBS; rename SBS' into SBS. 

  (* 2. Rely *)
  exploit rely_step; eauto.
    { inv ND'; unfold in_sb; simpl; rewrite (proj1 LR), res_plus_emp_l; ins; 
      try apply CC; try apply nodup_map; intros; rewrite ?in_map_iff in *;
      ins; desf; ins; vauto; try congruence;
        try (by try intro; red in CONS; desc; edestruct ACYC; eauto; vauto);
        try (eby destruct FRESH; eapply prefix_closed_hb); try done.
      by destruct H3; rewrite ?mhbE in *; desf; eauto using prefix_closed_hb. }
    by ins; eapply Isb; intro; desf. 
  clear V CC CO Irf Iesc Isb O LR; rename L into Lold; intros (L & ?); desf.
  generalize (fun a' IN => eq_trans (EsbS a') (SBS a' IN)); 
    clear EsbS SBS; intro SBS.
  assert (XXX: in_sb (b :: acts) Lold b = in_sb (b :: acts) L b).
    by clear - Esb ; unfold in_sb; induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear Lold XXX Esb. 

  (* 3. Ghost move *)
  eapply rely_compat_post in CC; eauto.
  2: intros c HB; assert (In c (b::acts)) by eauto using hb_in_acts2.
  2: ins; desf; [by red in CONS; desc; edestruct ACYC; eauto|].
  2: by destruct FRESH; eapply prefix_closed_hb; eauto.
  edestruct STEP0 with (rF := res_emp) as (P & GM & FUT); clear STEP0; eauto.
    eby rewrite res_plus_emp_r. 
  edestruct ghost_step with (I := @nil (escrowTy * resource))
    as [L' ?]; try rewrite res_lt_strip_l; unfold res_lt; 
    eauto using res_plus_emp_l, res_plus_emp_r; ins.

    apply res_plus_eq_empD in OUT; desc.
    apply res_plus_eq_empD in OUT0; desc.
    unfold incoming; rewrite INesc, OUT1, !res_plus_emp_r.
    eby eapply valid_ghost_refl, rely_def.

    unfold incoming; rewrite INesc, res_plus_emp_r.
    by eauto using res_plus_emp_r.

    apply LR; intro HB.
    destruct (Consistent_hb_in CONS HB) as [_ []]; desf. 
      by red in CONS; desc; edestruct ACYC; eauto.
    eby destruct FRESH; eapply prefix_closed_hb.

  rename L into Lold, L' into L; clear V CC CO INesc LR; desf.
  generalize (fun a' IN => eq_trans (EsbS a') (SBS a' IN)); 
    clear SBS; intro SBS.
  eapply res_plus_eq_empD in OUT; desc.
  eapply res_plus_eq_empD, proj1 in OUT0. 
  assert (XXX: in_sb (b :: acts) Lold b = in_sb (b :: acts) L b).
    by clear - Esb ; unfold in_sb; induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear XXX. 
  assert (XXX: out_sb (b :: acts) Lold b = out_sb (b :: acts) L b).
    by clear - Esb EsbS ; unfold out_sb; f_equal; auto;
       induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear Esb EsbS XXX. 
  assert (XXX: in_rf (b :: acts) Lold b = in_rf (b :: acts) L b).
    by clear - Erf ; unfold in_rf; induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear XXX. 
  assert (XXX: out_rf (b :: acts) Lold b = out_rf (b :: acts) L b).
    by clear - Erf ErfS ; unfold out_rf; f_equal; auto;
       induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear Erf ErfS XXX Lold.

  (* 4. Write protocol equivalence *)
  assert (WPE: pre_write_prot_equiv acts amap mo L b
               (in_sb (b :: acts) L b +!+ in_rf (b :: acts) L b) 
               (incoming (b :: acts) L b)).
    by eapply pwpe; eauto.
 
  specialize (FUT _ IN' (wpe _ _ _ _ _ _ _ WPE VG RS)); desf; rewrite res_plus_emp_r in *.
  (* 5. Guarantee *)
  edestruct guar_step as [L' ?]; eauto.
  rename L into Lold, L' into L.
  clear CC CO V VES VI VG FUT; desf.
  generalize (fun a' IN NEQ => eq_trans (EsbS a' NEQ) (SBS a' IN)); 
    clear SBS; intro SBS.

  eexists L, (l1' ++ (b,e', L (TsbS b), QQ) :: l2'),;
    [|split; [|exists nil; rewrite <- app_nil_end]]; rewrite ?map_app; try done.

  unfold gsafe, NW; repeat (apply conj; try done).
  ins; desf; rewrite filter_app; ins; desf; desf. 
  by rewrite !filter_eq_nil; ins; auto using res_plus_emp_r; case eqP; ins;
     destruct FRESH; desf; apply CT; eauto using in_or_app, in_cons.
  by rewrite <- filter_app; apply SBS; congruence.
  
  by intro; rewrite in_app_iff; ins; desf; eauto 6 using in_or_app, in_cons.
  
  by intros until 0; repeat (rewrite in_app_iff; simpl); intro; desf; eauto; 
     eapply safe_mon; try eapply SAFE; eauto using in_or_app, in_eq, in_cons with arith.

Case "fork".
  eapply map_eq_app_inv in THR; desf.
  destruct l2' as [|[[? r] QQ] l2']; ins; desf; ins.
  red in H; desf.
  exploit SAFE; eauto using in_or_app, in_eq.
  rewrite <- plus_n_Sm, <- plus_n_O; ins; desf.
  clear STEP END; exploit FORK; eauto; intro; desf.
  eexists L, (l1' ++ (a,_,r1,QQ) :: l2' ++ (a,_,r2,fun _ => Rtrue) :: nil),;ins. 
  2: do 3 (rewrite map_app; ins); split; vauto.
  2: by eexists; rewrite !map_app, ?app_ass; ins.
  red; unfold NW; repeat (split; try done).
    by ins; rewrite SBS; ins;
       repeat (rewrite ?filter_app, ?map_app, ?rsum_app; ins; desf); rewrite ?res_plus_emp_r;
       f_equal; rewrite res_plusC, res_plusAC.
    intro; repeat (rewrite in_app_iff; simpl); ins; desf; ins; 
    eauto 10 using in_or_app, in_eq, in_cons;
    apply (CT _ (in_or_app _ _ _ (or_intror _ (in_eq _ _)))).

  rewrite <- plus_n_Sm, <- plus_n_O in SAFE.
  intros until 0; repeat (rewrite in_app_iff; simpl); intro; desf; eauto; 
  eapply safe_mon; try eapply SAFE; eauto using in_or_app, in_eq, in_cons.
Qed.


Lemma gsafe_memsafe n acts amap sb rf mo L T :
  gsafe n acts amap sb rf mo L T ->
  mem_safe acts amap sb rf.
Proof.
  unfold gsafe; red; ins; desf; clear SAFE SBS CT.
  assert (GI: res_get (incoming acts L a) (loc (amap a)) <> p_bot).
    eapply V in IN; red in IN; desf.
    intro IN; eapply vghost_get_bot in VG; eauto.
    destruct (amap a); ins; desf; desf; try congruence.
    red in AG; desf; congruence.
  eapply visible_allocation in GI; eauto using Consistent_prefix_closed, prefix_closed_pres_in.
  desf; eauto 8.
Qed.

Lemma gsafe_racefree n acts amap sb rf mo L T :
  gsafe n acts amap sb rf mo L T ->
  race_free acts amap sb rf.
Proof.
  unfold gsafe; ins; desf; clear SAFE SBS CT.
   cut (forall a b
          (INa : In a acts)
          (INb : In b acts)
          (NEQ : a <> b)
          (LOC : loc (amap a) = loc (amap b)),
          is_access (amap a) ->
          is_access (amap b) ->
          is_na (amap a) ->
          happens_before amap sb rf a b \/ happens_before amap sb rf b a); ins.
     by red; ins; clear WRI; desf; [|rewrite or_comm]; eauto.
   apply NNPP; intro X; apply not_or_and in X; desc.

    assert (GI: res_get (incoming acts L b) (loc (amap b)) <> p_bot).
      eapply V in INb; red in INb; desf.
      intro INb; eapply vghost_get_bot in VG; eauto.
      destruct (amap b); ins; desf; desf; try congruence.
      red in AG; desf; congruence.

    assert (GI': res_get (incoming acts L a) (loc (amap a)) = p_uninit \/
                 exists v, res_get (incoming acts L a) (loc (amap a)) = p_na v).
      eapply V in INa; red in INa; desf; eauto.
      destruct (amap a); ins; desf; desf; 
        eauto using vghost_get_back_uninit;
        right; red in VG; desf; rewrite VG in *; ins;
      match goal with H : res_get ?r ?loc = p_na _ |- _ =>
        destruct r; ins; desf; inv_lplus loc; rewrite H in *; ins; desf; eauto
      end.
      eapply incoming_allocated in GI; desf.
      eapply incoming_uninit in GI'; desf.
      eapply compat_def2 with (edge := te) (edge' := te0); eauto; ins; desf;
        try (by intro Z; rewrite Z in *; ins; desf).
  by eapply hb_one in GI'; eauto; try (by intro Z; rewrite Z in *; ins; desf);
     eauto using mhb_hb_trans.
  by eapply hb_one in GI; eauto; try (by intro Z; rewrite Z in *; ins; desf);
     eauto using mhb_hb_trans.
  by destruct (L te), (L te0); ins; desf; inv_lplus (loc (amap a)); rewrite LOC in *;
     unfold prot_plus in *; desf.

      eapply incoming_na in GI'; desf.
      eapply compat_def2 with (edge := te) (edge' := te0); eauto; ins; desf;
        try (by intro Z; rewrite Z in *; ins; desf).
  by eapply hb_one in GI'; eauto; try (by intro Z; rewrite Z in *; ins; desf);
     eauto using mhb_hb_trans.
  by eapply hb_one in GI; eauto; try (by intro Z; rewrite Z in *; ins; desf);
     eauto using mhb_hb_trans.
  by destruct (L te), (L te0); ins; desf; inv_lplus (loc (amap a)); rewrite LOC in *;
     unfold prot_plus in *; desf.
Qed.


Lemma gsafe_rinit n acts amap sb rf mo L T :
  gsafe n acts amap sb rf mo L T ->
  initialized_reads acts amap rf.
Proof.
  unfold gsafe; red; ins; desf; clear SAFE SBS CT.
  assert (K:=IN); eapply V in K; red in K; desf.
  destruct (amap a) eqn: LEQ; ins; desf; desf.
  {
    assert (GI': res_get (incoming acts L a) l = p_na v0).
      red in VG; desf; rewrite VG in *; ins;
      match goal with H : res_get ?r ?loc = p_na _ |- _ =>
        destruct r; ins; desf; inv_lplus loc; rewrite H in *; ins; desf; eauto
      end.
    eapply visible_na in GI'; eauto using Consistent_prefix_closed, prefix_closed_pres_in; desf.
    red in CONS; desc; specialize (Crf a); intro Z; rewrite Z in Crf.
    by eapply Crf; eauto; rewrite LEQ; ins.
  }
  {
    eapply vghost_get_back_at in VG; eauto.
    eapply visible_atomic in VG; eauto using Consistent_prefix_closed, prefix_closed_pres_in; desf.
    red in CONS; desc; specialize (Crf a); intro Z; rewrite Z in Crf.
    by eapply Crf; eauto; rewrite LEQ; ins.
  }
  {
    eapply vghost_get_back_at in VG; eauto.
    eapply visible_atomic in VG; eauto using Consistent_prefix_closed, prefix_closed_pres_in; desf.
    red in CONS; desc; specialize (Crf a); intro Z; rewrite Z in Crf.
    by eapply Crf; eauto; rewrite LEQ; ins.
  }
Qed.


Lemma Consistent_ext_skip :
  forall acts amap sb rf mo
    (CONS : Consistent acts amap sb rf mo)
    a' (INa : In a' acts)
    b' (NIN : ~ In b' acts),
  Consistent (b' :: acts) amap  (fun a b => sb a b \/ a = a' /\ b = b') rf mo.
Proof.
  ins.
  assert (HBM: 
        forall a b,
          happens_before amap sb rf a b ->
          happens_before amap (fun a0 b => sb a0 b \/ a0 = a' /\ b = b') rf a b).
    by clear; induction 1; try red in H; desf; eauto using hb_sb, hb_trans.
  assert (HBE: forall a b,
       happens_before amap (fun a b => sb a b \/ a = a' /\ b = b') rf a b
       <-> happens_before amap sb rf a b \/
           a = a' /\ b = b' \/
           happens_before amap sb rf a a' /\ b = b').
    split; ins; desf; eauto using hb_sb.
    2: by eapply hb_trans, hb_sb; eauto.
    induction H; try red in H; desf; eauto using hb_trans;
    by eapply Consistent_hb_in in IHclos_trans2; eauto; desf.

  unfold Consistent, ExecutionFinite, ConsistentMO, restr_subset,
    ConsistentRF_basic, CoherentRR, CoherentRW, CoherentWR in *; unfold NW; desf. 
  repeat apply conj; ins; desf; eauto.

  by eapply CLOsb in H; desf; eauto.
  {
    intros x C; apply (ACYC x).
    destruct (eqP x b'); desf.
      exfalso; revert C; generalize b' at 3; ins; revert NIN.
      induction C; eauto; desf.
      eapply CLOsb in H; ins; desf.
      specialize (Crf y); desf; desf; ins.
      by destruct NIN; apply CLOlab; intro X; rewrite X in *.
    revert C n; clear - INa NIN Crf CLOlab CLOsb Crf.
    generalize x at 1 4; induction 1; desf; eauto using clos_trans.
    cut (y <> b'); eauto using clos_trans.
    intro; desf; clear - INa NIN C2 Crf CLOlab CLOsb.
    revert C2 NIN; generalize b' at 2 3; ins; revert NIN.
    induction C2; eauto; desf.
    eapply CLOsb in H; ins; desf.
    specialize (Crf y); desf; desf; ins.
    by destruct NIN; apply CLOlab; intro X; rewrite X in *.
  }
  by eapply Cmo3; eauto; rewrite HBE in *; desf; ins; 
     destruct NIN; apply CLOlab; intro X; rewrite X in *.
  by specialize (Crf a); desf; desf; ins.
  by specialize (Crf a); desf; desf; ins;
     rewrite HBE in *; desf; ins; eauto; 
     destruct NIN; apply CLOlab; intro X; rewrite X in *.
  by rewrite HBE in *; desf; ins; eauto;
     specialize (Crf b'); desf; desf; ins; 
     destruct NIN; apply CLOlab; intro X; rewrite X in *.
  by rewrite HBE in *; desf; ins; eauto;
     specialize (Crf b'); desf; desf; ins; 
     destruct NIN; apply CLOlab; intro X; rewrite X in *.
  by rewrite HBE in *; desf; ins; eauto; eapply Cmo in MO; desf; 
     destruct NIN; apply CLOlab; intro X; rewrite X in *.
Qed.

Lemma gsafe_final_sat :
  forall acts amap sb rf mo
    (CONS' : Consistent acts amap sb rf mo) L' 
    (V : forall a, In a acts -> valid_node acts acts amap rf L' a)
    (CC : compat amap sb rf L')
    (CO : conform acts amap mo L' acts) a r r_rem
    (SBS : L' (TsbS a) = r +!+ r_rem)
    (CT : In a acts) QQ
    (END : ghost_move r QQ)
    (ND : NoDup acts)
    (LR : label_restr acts amap sb rf L'),
  exists r, QQ r.
Proof. 
  ins; assert (FRESH: exists b, ~ In b acts); desf.
    exists (S (fold_right (fun x y => x + y) 0 acts)); clear.
    assert (forall n, In n acts -> n <= fold_right (fun x y => x + y) 0 acts).
      induction acts; ins; desf; [omega|].
      by eapply IHacts in H; omega.
    by intro X; eapply H in X; omega. 
  assert (ND' : NoDup (b::acts)) by vauto.
  assert (SKIP : amap b = Askip).  
    by red in CONS'; desc; apply NNPP; intro X; eapply FIN in X.
  assert (CONS: Consistent (b :: acts) amap (fun a0 b0 => sb a0 b0 \/ a0=a /\ b0=b) rf mo).
    by eapply Consistent_ext_skip; eauto.
  assert (PI  := Consistent_pres_in CONS).
  exploit prepare_step; eauto; vauto.
  clear CONS'.
  clear LR CC CO V; rename L' into Lold; intros (L & ?); desf.
  clear Lold LEQ SBS.
  exploit rely_step; eauto; vauto.
    { unfold in_sb; simpl; rewrite (proj1 LR), res_plus_emp_l; ins; 
      try apply CC; try apply nodup_map; intros; rewrite ?in_map_iff in *;
      ins; desf; ins; vauto; try congruence;
        try (by try intro; red in CONS; desc; edestruct ACYC; eauto; vauto);
        try (eby destruct FRESH; eapply prefix_closed_hb); try done.
      by rewrite ?mhbE in *; desf; eauto using prefix_closed_hb. }
    by ins; eapply Isb; intro; desf; tauto.

  clear V CC CO Irf Iesc Isb O LR; rename L into Lold; intros (L & ?); desf.
  assert (XXX: in_sb (b :: acts) Lold b = in_sb (b :: acts) L b).
    by clear - Esb ; unfold in_sb; induction (b :: acts); ins; f_equal; auto.
  rewrite XXX in *; clear Lold XXX Esb EsbS. 

  eapply rely_compat_post in CC; eauto; vauto.
  2: intros c HB; assert (In c (b::acts)) by eauto using hb_in_acts2.
  2: ins; desf; [by red in CONS; desc; edestruct ACYC; eauto|].
  2: by destruct FRESH; eapply prefix_closed_hb; eauto.
  rewrite SKIP in *; ins; desf.
  rewrite RELY in *.
  edestruct ghost_step with (I := @nil (escrowTy * resource))
    as [L' ?]; try rewrite res_lt_strip_l; unfold res_lt; 
    eauto using res_plus_emp_l, res_plus_emp_r; ins; desf; vauto.

    apply res_plus_eq_empD in OUT; desc.
    apply res_plus_eq_empD in OUT0; desc.
    unfold incoming; rewrite INesc, OUT1, !res_plus_emp_r.
    eby rewrite RELY in *; eapply valid_ghost_refl.

    unfold incoming; rewrite INesc, res_plus_emp_r, RELY.
    by eauto using res_plus_emp_r.

    apply LR; intro HB.
    destruct (Consistent_hb_in CONS HB) as [_ []]; desf. 
      by red in CONS; desc; edestruct ACYC; eauto.
    eby destruct FRESH; eapply prefix_closed_hb.
Qed.

(** Finally, we show our main theorem, namely adequacy. *)

Theorem adequacy : 
  forall e QQ (T: triple Rtrue e QQ) e' acts amap sb rf mo,
    cexecutions e e' acts amap sb rf mo ->
    mem_safe acts amap sb rf /\ 
    race_free acts amap sb rf /\ 
    initialized_reads acts amap rf /\
    forall v, e' = Evalue v -> exists r, (QQ v) r.
Proof.
  ins; red in T; red in H; desf.
  cut (forall n, exists L' T',
         gsafe n acts amap sb rf mo L' T' /\
         map erase T' = (a, e') :: l /\
         exists rem, map snd T' = QQ :: rem).
    clear; ins; specialize (H 1); desf. 
    repeat apply conj; eauto 6 using gsafe_rinit, gsafe_memsafe, gsafe_racefree.
    ins; desf.
    destruct T' as [|[[]]]; ins; desf.
    red in H; desf.
    specialize (CT _ (in_eq _ _)).
    specialize (SAFE _ _ _ _ (in_eq _ _)).
    specialize (SBS _ CT); ins; desf; desf; ins.
    by eapply gsafe_final_sat; eauto.

  remember ((a, e') :: l) as tmap; clear Heqtmap.
  remember {|
        Mtmap := tmap;
        Macts := acts;
        Mamap := amap;
        Msb := sb;
        Mrf := rf;
        Mmo := mo |} as c; revert tmap acts amap sb rf mo Heqc.
  apply clos_rt_rtn1 in H; induction H; ins; desf.
  2: by destruct y; exploit (IHclos_refl_trans_n1); ins; desf;
        eapply gsafe_pres in H; ins; eauto; desf; eauto;
        rewrite x2 in *; ins; eauto 8.

  assert (CONS: Consistent (init :: nil) (fun _ => Askip) (fun _ _ => False)
                     (fun _ => None) (fun _ _ => False)).
    red; unfold NW; repeat (split; ins); repeat red; ins.
    by induction H; ins; desf.

  exploit (T res_emp); clear T; try done; intro T.
  eapply ghost_step with (r:=res_emp) (I:=nil) (L:= fun _ => res_emp) in T; try exact CONS; eauto;
    try (solve [repeat split; ins]); rewrite ?res_plus_emp_l, ?res_plus_emp_r; 
    ins; eauto using res_lt_refl.
    2: by red; ins; rewrite rsum_eq_empI; ins; rewrite in_map_iff in *; desf; unfold upd; desf.
    2: by unfold incoming, out_esc, in_esc, in_rf, in_sb; rewrite res_plus_emp_l;
          repeat first [rewrite res_plus_eq_empI | rewrite rsum_eq_empI]; ins;
          repeat eexists; intros; desf; desf.
    2: by unfold incoming, out_esc, in_esc, in_rf, in_sb;
          repeat first [rewrite res_plus_eq_empI | rewrite rsum_eq_empI | 
            rewrite res_strip_emp]; ins; desf; apply res_lt_refl.
  desf; clear V.
  exists (upd L' (TsbS init) r), ((init, e, r, QQ)::nil),;ins; eauto.
  red; unfold NW.
  do 8 try (split; try done); ins; desf; rupd; ins; 
    eauto using in_eq, res_plus_emp_r, compat_guar1.
  red; unfold NW; exists r, I.
  unfold incoming, out_esc, in_sb, in_rf, in_esc, out_sb, out_rf in *; ins.
  rewrite escr_ffun_upd_irr, !res_plus_emp_r; rupd; try done.
  rewrite Esb, Erf, ErfS, ?res_plus_emp_l, ?res_plus_emp_r in *.
  repeat (split; try edone); ins.
  by destruct P'; intro; desf.
Qed.
