(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsModel GpsModelLemmas GpsGlobal GpsVisible.

Local Open Scope string_scope.

Set Implicit Arguments.

(******************************************************************************)
(** * 3.3.4. Protocol equivalence for writes *)
(******************************************************************************)

Definition pre_write_prot_equiv acts amap (mo : relation actid) L b r_pre r' :=
  match amap b with
    | Aalloc l n =>
          forall loc, l <= loc <= l + n -> res_get r' loc = p_bot
    | Astore ATat l _ =>
        forall tau S,
          res_get r_pre l <> p_bot ->
          res_get r' l = p_at tau S ->
          exists a, 
            << INa : In a acts >> /\
            << MO : immediate mo a b >> /\
            << Wa : is_writeL (amap a) l >> /\
          exists v_env, 
            (exists Sxxx, res_get (out_rf (b::acts) L a) l = p_at tau Sxxx /\ 
                          slist_lt S Sxxx) /\
            env_move r_pre l v_env (out_rf (b::acts) L a)
    | Armw l _ _ =>
        forall tau S,
          res_get r_pre l = p_at tau S ->
          exists Su,
            res_get r' l = p_at tau Su /\ get_state S = get_state Su
    | _ => True
  end.

Lemma pwpe :
  forall b acts amap sb rf mo
    (CONS : Consistent (b :: acts) amap sb rf mo)
    (ND' : NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (PI : pres_in sb rf b acts)
    L
    (RELY : rely (in_sb (b :: acts) L b) (amap b)
      (in_sb (b :: acts) L b +!+ in_rf (b :: acts) L b))
    I r
    (V : forall a : actid, In a acts -> valid_node (b :: acts) (b::acts) amap rf L a)
    (CC : compat amap sb rf (upd L (TescS b) (L (TescS b) +!+ r)))
    (CO : conform (b :: acts) amap mo L acts)
    (VG : valid_ghost (incoming (b :: acts) L b)
      (r +!+ out_esc (b :: acts) L b +!+ L (TcondS b)) I)
  (RS : res_lt (res_strip (incoming (b :: acts) L b)) r)
  (LR : label_restr (b :: acts) amap sb rf L),
   pre_write_prot_equiv acts amap mo L b 
     (in_sb (b :: acts) L b +!+ in_rf (b :: acts) L b) (incoming (b :: acts) L b).
Proof. 
  red; ins; destruct (amap b) eqn: W; ins; desf; desf; ins.

Case "Aalloc".
  apply NNPP; intro IN.
  exploit visible_allocation; eauto; intro M; desf.
  red in CONS; desc; destruct (Ca _ _ _ M0 _ _ _ W); clarify; try omega.
  by edestruct ACYC; eauto. 

Case "Astore ATat".
  assert (GG := vghost_get_at _ VG RS _ H0); clear H0.

  assert (RG: exists S_pre, res_get (in_sb (b :: acts) L b +!+ in_rf (b :: acts) L b) l = p_at tau S_pre). 
    eby eapply vghost_get_back; eauto; rewrite res_plusA.
  clear H; desf.

  edestruct res_get_plus with 
    (r := in_sb (b :: acts) L b +!+ in_rf (b :: acts) L b) 
    (r' := in_esc (b :: acts) L b) as (S' & EQ & SUB); eauto.
    unfold incoming in *; rewrite res_plusA; red in VG; desc; congruence.

  unfold incoming in VG.
  rewrite res_plusA in *.
  exploit vghost_get_at; eauto.
  intro Z; rewrite Z in *; desf.

  exploit exists_immediate_helper; (try solve [clear CC; eauto]); eauto.
    by rewrite W.
  intro AW; desf.
  unfold NW in *. 
  eexists,;[|split; [|split]]; eauto.

  inv ND'.
  desf; generalize (V _ INa); intro V'; red in V'; desf.
  assert (DEF: in_sb (b :: acts) L b +!+ L (TrfS a) <> Rundef).
    by eapply helper_plus_insb_rfS_def; eauto.
  

  assert (K := res_plus_get1 _ _ l DEF).

  unfold env_move, NW.

  assert (AG: exists v0, atomic_guar r0 l v0 (out_sb (b :: acts) L a)
                                            (out_rf (b :: acts) L a)).
    by red in GUAR; desf; desf; ins; desf; vauto.


  desf; generalize AG; red in AG; desc; intros; desf;
  (*eexists (out_rf (b :: acts) L a);*) revert DEF K; rewrite RELY, RG, Grf in *;
  rewrite VrfS, Grf; ins; desf; eauto using no_rmw_between; try by rewrite W in *.

  SCase "initialization write".
    edestruct visible_atomic with (b:=b) (l:=l) as [c ?]; 
      try solve [clear CC; eauto]; eauto; desf.
    assert (In c acts) by eauto using pres_in_hb.

    assert (get_state S' = s) by (destruct S'; ins; desf).

    destruct (eqP c a) as [|NEQ]; [subst; rewrite Grf in *; desf|].

    do 4 try (eexists; try edone); repeat split; try edone.
    by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.

  red in CONS; desc; eapply Cmo in NEQ; desf; eauto.
  2: by eapply MO in NEQ; ins; eauto; eapply Cmo; eauto; rewrite W.

  eapply CO in NEQ; eauto.
  rewrite H4, Grf in *; ins; desf.

    do 4 try (eexists; try edone); repeat split; try edone.
      by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.
    by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.

  SCase "normal write".

  assert (get_state S' = s).
    by eapply get_state_eq_prove; ins; rewrite SEQ in *; desf; 
       eauto using prot_trans_refl.

  edestruct visible_atomic with (b:=b) (l:=l) as [c ?]; 
      try solve [clear CC; eauto]; eauto; desf.
  (*rewrite <- res_plusA, RELY in EQ; rewrite EQ in *; ins; desf.*)
    assert (In c acts) by eauto using pres_in_hb.

    destruct (eqP c a) as [|MO']; [subst; rewrite Grf in *; desf|].
  do 4 try (eexists; try edone); repeat split; try edone.
  by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.
  

  red in CONS; desc; eapply Cmo in MO'; desf; eauto.
  2: by eapply MO in MO'; ins; eauto; eapply Cmo; eauto; rewrite W.

  eapply CO in MO'; eauto.
  rewrite H5, Grf in *; ins; desf.
  do 4 try (eexists; try edone); repeat split; try edone.
    by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.
  by eapply slist_lt_trans; eauto; red; eauto using prot_trans_refl.

Case "Armw".
  edestruct res_get_plus with (r' := in_esc (b :: acts) L b) as (S' & EQ & SUB); eauto.
  unfold incoming in *; rewrite res_plusA; red in VG; desc; congruence.
  rewrite res_plusA in *.
  exploit visible_atomic; try solve [clear CC; eauto]; eauto.
  intro M; desf.

  red in CONS; desc; red in Cmo; desc.
  specialize (Crf b).
  destruct (rf b) as [b_rf|] eqn: RF; desf.
  2: by edestruct Crf; eauto; rewrite W; ins.
  rewrite W in *; ins; desf.
  assert (INb: In b_rf acts) by (by eapply PI in RF).
  specialize (V _ INb); red in V; desf. 

  destruct (classic (a = b_rf)) as [|NEQ]; desf.

  SCase "equal".
    rewrite <- (VrfM b) in M2; auto using in_eq.
    2: by rewrite W.
    ins; desf; desf. 
    generalize H; unfold in_rf; rewrite res_plusC; intro.
    eapply res_get_sum with (l := map L (map (Trf^~ b) (b :: acts))) in M2; desf; eauto.
    2: by intro X; rewrite X in *. 
    2: by eapply in_map, in_map_iff; eauto using in_cons.
    eapply res_get_plus with (r' := in_sb (b :: acts) L b) in M2.
    2: by intro X; rewrite X in *. 
    desf; rewrite M2 in *; desf.


    eapply vghost_get_at in VG; try edone. 
    eexists,; eauto.
    eapply get_state_eq_incl_helper; eauto.
    eapply slist_lt_trans with S'; eauto.
      by red; eauto using prot_trans_refl.
    eapply slist_lt_trans with S'0; eauto.
    by red; eauto using prot_trans_refl.

  SCase "non-equal".
    eapply Cmo1 in NEQ; desf; eauto.
    2: by edestruct Cwr; eauto; instantiate;
          rewrite W; ins; destruct (amap a).
    eapply CO in NEQ; eauto using pres_in_hb. 
    rewrite M2, <- (VrfM b) in NEQ; auto using in_eq.
    2: by rewrite W.
    ins; desf; desf. 
    generalize H; unfold in_rf; rewrite res_plusC; intro.
    eapply res_get_sum with (l := map L (map (Trf^~ b) (b :: acts))) in Heq; desf; eauto.
    2: by intro X; rewrite X in *. 
    2: by eapply in_map, in_map_iff; eauto using in_cons.
    eapply res_get_plus with (r' := in_sb (b :: acts) L b) in Heq.
    2: by intro X; rewrite X in *. 
    desf; rewrite Heq in *; desf.
   
    eapply vghost_get_at in VG; try edone. 
    eexists,; eauto.
    eapply get_state_eq_incl_helper; eauto.
    eapply slist_lt_trans with S'; eauto.
      by red; eauto using prot_trans_refl.
    eapply slist_lt_trans with S'0; eauto.
    eapply slist_lt_trans with s; eauto.
    by red; eauto using prot_trans_refl.
Qed.

Lemma wpe :
  forall b acts amap mo L r_pre r r' r'' I,
    pre_write_prot_equiv acts amap mo L b r_pre r ->
    valid_ghost r (r' +!+ r'') I ->
    res_lt (res_strip r) r' ->
    write_prot_equiv (amap b) r_pre r'.
Proof.
  unfold pre_write_prot_equiv, write_prot_equiv; ins; desf; ins.
  Case "Aalloc".
    exploit H; eauto; clear - H0; red in H0; desf; ins.
    destruct r'; ins; desf; inv_lplus loc. 
    by rewrite x0 in *; unfold prot_plus in *; desf.
  Case "Awrite".
    exploit H; clear H; eauto using vghost_get_back_at; try congruence.
    by intro; desf; eauto 10.
  Case "Armw".
    exploit H; eauto; ins; desf; eauto using vghost_get_at. 
Qed.

(******************************************************************************)
(** * 3.3.5. Guarantee *)
(******************************************************************************)

Lemma conform_suff_immediate :
  forall b acts amap sb rf mo 
    (CONS: Consistent (b::acts) amap sb rf mo)
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (PI: pres_in sb rf b acts) L
    (V : forall a (IN: In a acts), valid_node (b::acts) (b::acts) amap rf L a)
    (CO: conform (b::acts) amap mo L acts) r'
    (CC: compat amap sb rf (upd L (TescS b) (L (TescS b) +!+ r')))
    (NIN: ~ In b acts) loc
    (Orf: rsum (map L (map (Trf b) (b::acts))) = res_emp)
    (LR : label_restr (b :: acts) amap sb rf L)
    (LOC: is_atomic_writeL (amap b) loc)
    (HPL: res_get (incoming (b :: acts) L b) loc <> p_bot)
    (G : res_get r' loc <> p_bot) vb rsb r
    (MAG : atomic_guar r' loc vb rsb r)
    (PREV: forall a (IMM: immediate mo a b) (IN: In a acts)
      (AW: is_atomic_writeL (amap a) loc),
      prot_trans_ok (res_get (out_rf (b::acts) L a) loc) (res_get r loc))
    (GIMP: forall tau S
      (G1: res_get r loc = p_at tau S) S'
      (G2: res_get (L (TescS b)) loc = p_at tau S'),
      slist_lt S' S),
  conform (b::acts) amap mo (upd L (TrfS b) r) (b::acts).
Proof.
  ins; assert (PREV': forall a (IN: In a acts) (MO: mo a b)
                    (AW: is_atomic_writeL (amap a) loc),
          prot_trans_ok (res_get (out_rf (b::acts) L a) loc) (res_get r loc)).
  {
    ins; red in CONS; desc; red in Cmo; desc.
    destruct exists_immediate with (r:=mo) (a:=a) (b:=b) (l:=b::acts) 
      as (a' & IMM & DISJ); eauto.
      intros ? X; eapply Cmo in X; ins; desf.
      by eapply (proj1 FIN); clear - X0; intro Y; rewrite Y in *.
    assert (IN': In a' acts).
      exploit (proj1 FIN a'); intro; desf; try eapply Cmo in DISJ; try rewrite H in *; ins; desf.
      by exfalso; destruct IMM; eauto.
    assert (W' : is_writeL (amap a') loc).
      by desf; eauto; eapply proj1, Cmo in IMM; desf; destruct (amap b); ins; desf; desf.
    desf; [|eapply prot_trans_ok_trans; [eapply CO|]]; try eapply (PREV a'); eauto.
    eapply CO in DISJ; eauto.
    specialize (V _ IN'); red in V; desc.
    destruct (amap a'); ins; desf; desc.
    rewrite GUAR0 in DISJ; clear - DISJ; red in DISJ; ins; desf.
  }  
  clear PREV; red; ins; unfold out_rf; rewrite !map_upd_irr;
    try (by rewrite in_map_iff; intro; desf).
  ins; desf; desf; rupd; try congruence; try (eby eapply CO); rewrite ?Orf, ?res_plus_emp_r.
    by exfalso; red in CONS; desc; red in Cmo; desf; eauto.

  assert (loc = loc0) by (destruct (amap b0); ins; desf); subst. 
  by eapply PREV'; eauto.


  (* AFTER *)

  assert (CONS' := CONS); red in CONS'; desc; red in Cmo; desc.
  assert (loc0 = loc) by (destruct (amap a); ins; desf; desf); subst.
  edestruct exists_immediate with (r:=(fun x y => mo y x)) (l:=a::acts) 
    as (a' & [N N'] & DISJ); unfold transitive; eauto. 
    intros ? _ X; eapply Cmo in X; desf.
    by eapply (proj1 FIN); clear - X0; intro Y; rewrite Y in *.
  assert (IN': In a' acts).
    exploit (proj1 FIN a'); intro; desf; try eapply Cmo in DISJ; try rewrite H in *; ins; desf.
    by exfalso; eauto.
  assert (W' : is_writeL (amap a') loc).
    by desf; eauto; eapply Cmo in N; desf; destruct (amap a); ins; desf; desf.

  assert (AW: is_atomic_writeL (amap a') loc). {
    clear DISJ b0 INb Wb MO.
    assert (V' := V _ IN'); red in V'; desf.
    destruct (amap a'); ins; desf.
    exploit helper_plus_inc_escS_back_def;
      eauto using in_eq, in_cons; ins.
    rewrite incoming_upd_irr, upds; ins.
    rewrite res_plusAC, res_plusC.
    replace (incoming (a :: acts) L a' +!+ r') with Rundef; try done.
    red in VG; desf; rewrite VG; ins; desf;
    destruct r0; ins; desf; inv_lplus loc;
    rewrite GUAR1 in *; ins; desf; rewrite <- H0 in *; ins; desf.
  }
  destruct (is_rmw (amap a')) eqn: RMW. {
    revert DISJ; specialize (Crf a'); desf.
    exploit Crmw; eauto; []; intros [? ?].
    intro; exfalso; destruct (eqP a a0) as [|NEQ]; revert DISJ; desf; [by apply PC in Heq|]. 
    edestruct (Cmo _ _ H) as (? & ? & ?).
    ins; desf; desf; destruct (amap a'); ins; desf; desf; subst;
      try by eapply Cmo1 in NEQ; desf; eauto.
    exfalso.
    assert (V' := V _ IN'); red in V'; desf.
    destruct (amap a'); ins; desf.
    eapply vghost_get_back_at in VG; eauto.
    clear PI;
    exploit visible_atomic; try solve [clear CC; eauto]; eauto.
      by red in PC; desf; split; ins; eauto.
    by intro; desf; eauto.
  }
  cut (prot_trans_ok (res_get r loc) (res_get (out_rf (a::acts) L a') loc)).
    by ins; desf; eapply prot_trans_ok_trans, CO; eauto.
  clear b0 DISJ INb MO Wb.
  assert (V' := V _ IN'); red in V'; desf.

  assert (DEF := helper_plus_inc_escS_back_def _ _ CONS ND CC N (lr_upd_escS _ _ LR (in_eq _ _))).
  rewrite incoming_upd_irr, upds in DEF; ins.

  destruct (amap a') eqn: LEQ'; ins; desf; unfold atomic_guar in GUAR; desf.
    exfalso.
    exploit visible_uninit2; try solve [clear PI; eauto using vghost_get_back_uninit];
      eauto using prefix_closed_pres_in. 
    rewrite mhbE; intro; desf.
    by eapply Cmo3 in H; eauto; instantiate; rewrite LEQ'.
  exploit vghost_get_back_at; eauto; intro Gin; desf.

  revert MAG; intro. 
  red in MAG; desf.
    destruct DEF.
    rewrite res_plusAC, res_plusC.
    destruct (incoming (a :: acts) L a'); ins; desf.
    exfalso; inv_lplus loc; ins.
    by red in VG; desf; unfold prot_plus in *; desf.

  assert (get_state S'0 = s0); subst.
    by eapply get_state_eq_prove; ins; rewrite SEQ0 in *; 
       desf; eauto using prot_trans_refl.

  eapply WG; red; unfold NW; repeat (eexists; try edone).
  {
    intro X; apply DEF; unfold incoming.
    apply (f_equal (fun x => x +!+ L (TescS a) +!+ in_esc (a :: acts) L a' +!+ rsb)) in X.
    revert X; simpl; rewrite !res_plusA, !(res_plusAC _ r), (res_plusC r), REQ0. 
    rewrite !(res_plusAC _ (L (TescS a))). 
    rewrite <- !res_plusA; intro X; rewrite res_plusC in X |- *.
    eapply plus_upd_def; eauto.
    ins; rewrite SEQ0 in *; desf; clear X.
    eapply get_plus_at_oneD in G'; eauto; desf.
    2: by edestruct GIMP as (?&?&?); eauto using get_state_lt, prot_trans_trans.
   
    rewrite res_plusA in G'.
    change (in_sb (a :: acts) L a' +!+
            in_rf (a :: acts) L a' +!+ in_esc (a :: acts) L a')
      with (incoming (a :: acts) L a') in *.

    exploit visible_atomic2_cor; try solve [clear CC; eauto]; 
      eauto using prefix_closed_pres_in.
    intro; desc; rewrite G' in *; desf.
    rewrite mhbE in *; desf; [by edestruct NEW; eauto|].

    right; destruct (classic (a = a0)) as [|NEQ]; desf.
      by destruct NIN; eauto using prefix_closed_hb.
    eapply Cmo1 in NEQ; eauto; desf.
      by exfalso; eapply Cmo3 in MHB; eauto; (try by rewrite LEQ').
    exploit (PREV' a0); eauto using prefix_closed_hb.
    by rewrite Gout, Grf0, <- slist_ltD; ins; desf.
  }

  clear PI; exploit visible_atomic; unfold incoming; try solve [clear CC; eauto]; eauto.
    by red in PC; desf; split; ins; eauto.
  intro; desf; eauto.
  assert (NEQ: a0 <> a).
    by intro; desf; destruct NIN; eauto using prefix_closed_hb.
  assert (is_writeL (amap a') loc).
    clear - ATa N Cmo; apply Cmo in N; desf.
    by destruct (amap a); ins; desf; desf.
  eapply Cmo1 in NEQ; eauto; desf. 
  2: by exfalso; eapply N' in NEQ; try eapply Cmo3; eauto.
  eapply PREV' in NEQ; eauto using prefix_closed_hb.
  unfold incoming in Gin; rewrite <- res_plusA in Gin. 
  destruct (res_get_plus_back _ _ _ Gin) as (?&AA&?); try done.
  rewrite x3, AA, Grf0 in *; ins; desf.
  split; ins; eapply slist_lt_trans; eauto.
  by red; eauto.
Qed.

Lemma vghost_get_back_at2 :
  forall r1 r r' I 
    (VG : valid_ghost r1 (r +!+ r') I) l tau S
    (G  : res_get r l = p_at tau S),
    exists S_pre, 
      res_get r1 l = p_at tau S_pre.
Proof.
  ins; red in VG; desf; clear I VG1. 
  destruct r; ins; desf; ins; desf.
  inv_lplus l; rewrite G in *; destruct (pmap l); ins; desf; eauto.
Qed.

Lemma guar_sidecond :
  forall r l v r_sb r_rf 
    (G : atomic_guar r l v r_sb r_rf)
    acts L b I
    (VG : valid_ghost (incoming acts L b)
           (r +!+ out_esc acts L b +!+ L (TcondS b)) I)
    (RS : res_lt (res_strip (incoming acts L b)) r) tau S 
    (G1 : res_get r_rf l = p_at tau S) S'
    (G2 : res_get (L (TescS b)) l = p_at tau S'),
    slist_lt S' S.
Proof.
   ins; unfold out_esc in *; ins; rewrite !res_plusA, res_plusAC in *.
   exploit vghost_get_back_at2; eauto; intro X; desf.
   red in VG; desf.
   rewrite VG in *; ins.
   edestruct res_get_plus 
     with (r':=r +!+ rsum (map L (map (Tesc b) acts)) +!+ L (TcondS b))
     as (?&?&?); eauto.
     congruence.
   rewrite VG0 in *; ins; rewrite X in *; desf.
   red in RS; desf.
   rewrite !res_plusA in *.
   red in G; desf.
     destruct (L (TescS b)); ins; desf.
     inv_lplus l; clear Heq2; ins.
     by rewrite X, GET in *; ins; desf.

   destruct (L (TescS b)); ins; desf.
   inv_lplus l; clear Heq2; ins.
   rewrite X, G1, G2, GET in *; ins; desf.
   by clear - H0 SEQ; red; ins; eexists; try rewrite SEQ; eauto using prot_trans_refl.
   by clear - H0 SEQ Heq7; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
   by clear - H0 SEQ Heq7; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
   by clear - H0 SEQ Heq8; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
Qed.

Lemma guar_sidecond2 :
  forall r l v r_sb r_rf 
    (G : atomic_guar r l v r_sb r_rf)
    acts L b I
    (VG : valid_ghost (incoming acts L b)
           (r +!+ out_esc acts L b +!+ L (TcondS b)) I)
    (RS : res_lt (res_strip (incoming acts L b)) r) tau S 
    (G1 : res_get r_rf l = p_at tau S) S'
    (G2 : res_get (L (TcondS b)) l = p_at tau S'),
    slist_lt S' S.
Proof.
   ins; rewrite <- ?res_plusA, res_plusC in *.
   exploit vghost_get_back_at2; eauto; intro X; desf.
   red in VG; desf.
   rewrite VG in *; ins.
   edestruct res_get_plus 
     with (r':=r +!+ out_esc acts L b)
     as (?&?&?); eauto.
     congruence.
   rewrite VG0 in *; ins; rewrite X in *; desf.
   red in RS; desf.
   rewrite !res_plusA in *.
   red in G; desf.
     destruct (L (TcondS b)); ins; desf.
     inv_lplus l; clear Heq2; ins.
     by rewrite X, GET in *; ins; desf.

   destruct (L (TcondS b)); ins; desf.
   inv_lplus l; clear Heq2; ins.
   rewrite X, G1, G2, GET in *; ins; desf.
   by clear - H0 SEQ; red; ins; eexists; try rewrite SEQ; eauto using prot_trans_refl.
   by clear - H0 SEQ Heq7; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
   by clear - H0 SEQ Heq7; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
   by clear - H0 SEQ Heq8; red; ins; eexists; try rewrite SEQ; 
      eauto 7 using prot_trans_refl, In_mk_slist, in_or_app.
Qed.



Lemma res_lt_rf_in acts a b L : 
  In a acts -> 
  NoDup acts -> 
  res_lt (L (Trf a b)) (incoming acts L b).
Proof.
  ins; unfold incoming, in_rf.
  eapply In_NoDup_Permutation in H; desf.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ H))); ins.
  eauto using res_lt_plus_l, res_lt_plus_r, res_lt_refl.
Qed. 


Lemma compat_atom_write :
  forall b acts amap sb rf mo 
    (CONS: Consistent (b::acts) amap sb rf mo)
    (PC: prefix_closed sb rf acts)
    (ND: NoDup (b::acts)) l
    (AW: is_atomic_writeL (amap b) l) M L
    (V : forall a, In a acts -> valid_node (b :: acts) M amap rf L a) r
    (CC : compat amap sb rf (upd L (TsbS b) r))
    (CO : conform (b :: acts) amap mo L acts) I
    (VG : valid_ghost (incoming (b :: acts) L b)
                      (r +!+ out_esc (b :: acts) L b +!+ L (TcondS b)) I)
    (Orf : L (TrfS b) = res_emp)
    (Orf0 : rsum (map L (map (Trf b) (b::acts))) = res_emp)
    (Osb : L (TsbS b) = res_emp)
    (Osb0 : rsum (map L (map (Tsb b) (b::acts))) = res_emp)
    (LR : label_restr (b :: acts) amap sb rf L)
    tau s S' r_sb r_rf
    (REQ : r_sb +!+ r_rf = res_upd r l (p_at tau S'))
    (Gsb : res_get r_sb l = p_at tau S')
    (Grf : res_get r_rf l = p_at tau S')
    S (GET : res_get r l = p_at tau S)
    v (AG : atomic_guar r l v r_sb r_rf)
    (LT: res_lt (res_strip (incoming (b :: acts) L b)) r)
    (LAST : forall s0, In s0 (SL_list S) -> prot_trans tau s0 s)
    (SEQ : forall s0, In s0 (SL_list S') <-> s0 = s \/ In s0 (SL_list S))
    (CO' : conform (b :: acts) amap mo (upd L (TrfS b) r_rf) (b :: acts)),
    compat amap sb rf (upd L (TsbS b) (res_upd r l (p_at tau S'))).
Proof.
  repeat red; ins; eapply CC; eauto.
  revert H.
  destruct (classic (In (TsbS b) E)) as [N|N];
   [apply In_NoDup_Permutation in N; desf;
    rewrite !(Permutation_rsum (Permutation_map _ N)); simpls |];
  rewrite !map_upd_irr; rupd; ins. 
  eapply plus_upd_def; eauto.
  intros.

  rewrite SEQ in *; desf.
  eapply get_rsum_at_oneD in G'; eauto; desf.
  rewrite in_map_iff in *; desf.
  assert (get_state S' = s); subst.
    by apply get_state_eq_prove; ins; rewrite SEQ in *; desf; eauto using prot_trans_refl.

  assert (IN' : In (te_src x) (b :: acts)).
    by eapply lab_src_in_acts; eauto.

  simpls; desf.
  revert G'0; intro.
  {
    assert (K:=VG); eapply vghost_get_back_at in K; eauto.

    destruct x; ins.
    rewrite (rsum_eq_empD (_ :: _) Osb0 (L (Tsb a b))) in G'0; try done.
    eapply (in_map L (_ :: _)), (in_map (Tsb a) (_ :: _)). 
    by eapply lab_tgt_in_acts; eauto.
        
    rewrite (rsum_eq_empD (_ :: _) Orf0 (L (Trf a b))) in G'0; try done.
    eapply (in_map L (_ :: _)), (in_map (Trf a) (_ :: _)). 
    by eapply lab_tgt_in_acts; eauto.
          
    by ins; rewrite Orf in G'0.

    cut (L (Tesc a b) = res_emp); [by intro XX; rewrite XX in *; ins|].
    apply LR; intro HB.
    destruct (hb_in_acts2 CONS HB); desf.
      by red in CONS; desc; edestruct ACYC; eauto.
    by inv ND; eauto using prefix_closed_hb.

    exploit guar_sidecond; eauto.
    by intro; desf; right; eauto using prot_trans_trans, get_state_lt.

    exploit guar_sidecond2; eauto.
    by intro; desf; right; eauto using prot_trans_trans, get_state_lt.
  }

  exploit visible_atomic2; try solve [clear CC; eauto]; eauto.
    eby eapply compat_mon_upd with (a := TsbS b); ins; rewrite Osb, res_plus_emp_l.
  intro; desf.
  destruct (eqP a b) as [|NEQ]; desf.
    by unfold out_rf in *; ins; rewrite Orf, Orf0, res_plus_emp_r in *; ins.
  assert (IN: In a (b::acts)).
    by rewrite mhbE in *; ins; desf; eauto using prefix_closed_hb, pres_in_hb.
  clear IN'; ins; desf.
  assert (MO:=NEQ).
  red in CONS; desc; red in Cmo; desc; eapply Cmo1 in MO; eauto; try by rewrite LEQ.
  desf.
    eapply CO' in MO; eauto using in_eq, in_cons; try (by rewrite LEQ).
    unfold out_rf in *; revert MO; simpls; rupd; try congruence.
    rewrite !map_upd_irr, Gout, Orf0, res_plus_emp_r, Grf; 
      try (by rewrite in_map_iff; intro; desf); ins; desf.
    by rewrite slist_ltD in *; eauto.
  eapply CO' in MO; eauto using in_eq, in_cons; try (by rewrite LEQ; ins; desf; desf).
  unfold out_rf in *; revert MO; simpls; rupd; try congruence.
  rewrite !map_upd_irr, Gout, Orf0, res_plus_emp_r, Grf; 
    try (by rewrite in_map_iff; intro; desf); ins; desf.
  by rewrite slist_ltD in *; eauto.
Qed.


Lemma vghost_get_bot r1 r r' I l : 
  valid_ghost r1 (r +!+ r') I ->
  res_get r1 l = p_bot ->
  res_get r l = p_bot.
Proof.
  unfold valid_ghost; ins; desf; destruct r; ins; desf.
  inv_lplus l; ins; unfold prot_plus in *; desf; congruence.
Qed.


Lemma guar_step :
  forall b acts (NIN: ~ In b acts) amap sb rf mo r
    (CONS: Consistent (b :: acts) amap sb rf mo) 
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (PI: pres_in sb rf b acts) r_pre r_sb r_rf
    (G: guar r_pre r (amap b) r_sb r_rf) L
    (RPRE: r_pre = in_sb (b::acts) L b +!+ in_rf (b::acts) L b) r_pre_sb
    (RELY: rely r_pre_sb (amap b) r_pre)
    (V : forall a (IN: In a acts), valid_node (b::acts) (b::acts) amap rf L a)
    (CC: compat amap sb rf (upd L (TescS b) (L (TescS b) +!+ r)))
    (CO: conform (b::acts) amap mo L acts) I 
    (VG: valid_ghost (incoming (b::acts) L b)
                     (r +!+ out_esc (b::acts) L b +!+ L (TcondS b)) I) 
    (VI : forall sigma r (GIN: In (sigma, r) I), 
            exists P P', interp_escr sigma = (P, P') /\ P' r)
    (Orf: out_rf (b::acts) L b = res_emp)
    (Osb: out_sb (b::acts) L b = res_emp)
    (RS:  res_lt (res_strip (incoming (b::acts) L b)) r)
    (VES: res_lt (rsum (map snd (filter (escr_ffun L (b::acts)) I))) (L (TescS b)))
    (LR : label_restr (b :: acts) amap sb rf L)
    (WPE: pre_write_prot_equiv acts amap mo L b r_pre (incoming (b::acts) L b)),
  exists L',
    << V : forall a (IN: In a (b::acts)), valid_node (b::acts) (b::acts) amap rf L' a >> /\
    << CC: compat amap sb rf L' >> /\
    << CO: conform (b::acts) amap mo L' (b :: acts) >> /\
    << EsbS: forall a, a <> b -> L' (TsbS a) = L (TsbS a) >> /\
    << EsbS': L' (TsbS b) = r_sb >> /\
    << LR : label_restr (b :: acts) amap sb rf L' >>.
Proof.
  intros; unfold NW.
  apply res_plus_eq_empD in Orf; desc.
  apply res_plus_eq_empD in Osb; desc.
  assert (NRF: forall c, ~ rf c = Some b).
    intros c RF; red in CONS; desc.
    specialize (Crf c); rewrite RF in *; desc.
    destruct (proj1 FIN c); desf. 
      by intro X; rewrite X in *. 
     by eapply ACYC; vauto.
    by eapply PC in RF.
  assert (NSW: forall c, ~ synchronizes_with amap rf b c).
    eby unfold synchronizes_with; red; ins; desc; edestruct NRF.
  assert (LRF: L (Trf b b) = res_emp) by (apply LR; done).
  destruct (amap b) eqn: LEQ; ins; desf.

Case "Askip".

  exists (upd L (TsbS b) r); repeat (split; eauto using in_eq, upds, compat_guar1); ins;
    rupd; desc; try congruence.
  destruct IN; clarify; [|by apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr; rupd; try done; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 

  by eapply conform_upd_irr, conform_ext_irr; eauto; rewrite LEQ; ins.

Case "Aalloc".

  assert (V': forall a (IN: b = a \/ In a acts),
    valid_node (b :: acts) (b :: acts) amap rf
      (upd L (TsbS b) (res_mupd r l n p_uninit)) a).
  {
    ins; destruct IN; clarify; [|by apply valid_node_upd_irr; eauto; intro; desf].
    red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
    unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
    rewrite ?map_upd_irr, ?escr_ffun_upd_irr; rupd; try done; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 
  }

  exists (upd L (TsbS b) (res_mupd r l n p_uninit)); 
    repeat (split; eauto using in_eq, upds, compat_guar1); ins;
    rupd; desc; try congruence.

  eapply compat_guar1 in CC; subst.
  red; ins; intro.
  eapply CC; eauto; revert H.
  destruct (classic (In (TsbS b) E)) as [P|];
    [eapply In_NoDup_Permutation in P; desc; clarify;
     rewrite !(Permutation_rsum (Permutation_map _ P)); try done |]; simpl;
    rewrite !map_upd_irr; try done; rupd; eauto.
    intro; apply NNPP; intro.
    assert (BOT: forall loc, l <= loc <= l + n -> res_get (r +!+ rsum (map L l')) loc = p_bot).
    { 
      ins; rewrite res_get_plus_bot, res_get_rsum_bot; try kundef.
      red in WPE; rewrite LEQ in WPE; specialize (WPE _ H1). 
      split; eauto using vghost_get_bot.
      ins; rewrite in_map_iff in *; desf.
      apply NNPP; intro M.
      assert (INx: In (te_src x) (b :: acts)). 
        by eapply lab_src_in_acts; eauto.
      edestruct visible_allocation_one with (N := b::acts) (te:=x) 
        (L := upd L (TsbS b) (res_mupd r l n p_uninit)) (l:=loc); 
        eauto using Consistent_prefix_closed, in_eq; desf.
        by rupd; intro; desf.

      assert (AA: L x <> res_emp) by (intro AA; rewrite AA in *; simpls).
      assert (BB := lab_tgt_in_acts CONS LR _ AA); clear AA.
      red in CONS; desc; destruct (Ca _ _ _ H5 _ _ _ LEQ); clarify; try omega.
      ins; desf. 
        clear - Orf Orf0 Osb Osb0 WPE M VG BB.
        assert (r +!+ out_esc (te_src x :: acts) L (te_src x) +!+ L (TcondS (te_src x)) <> Rundef).
          by red in VG; desf; congruence.
        rewrite res_plusC in VG; eapply vghost_get_bot in VG; eauto; unfold out_esc in *.
        do 2 (apply res_get_plus_bot in VG; desc; try kundef).
        destruct M, x; ins; rewrite ?Orf, ?Osb; ins; specialize (BB _ eq_refl).
        by erewrite rsum_eq_empD with (l := L (Tsb a a) :: _) (r := L (Tsb a b)); eauto;
           desf; ins; eauto using in_map.
        by erewrite rsum_eq_empD with (l := L (Trf a a) :: _) (r := L (Trf a b)); eauto;
           desf; ins; eauto using in_map.
        by eapply res_get_rsum_bot with (rl := _ :: _); eauto; ins; try kundef;
           desf; ins; eauto using in_map.
      by rewrite mhbE in *; destruct NIN; desf; eauto using prefix_closed_hb.
   }
   clear - BOT H H0.
   destruct r; ins; desf.
   eapply lift_plus_none_inv in Heq2; desc; inv_lplus l0; ins; desf; try congruence.
   desf; rewrite BOT in Heq0; auto; clear - Heq0 Heq2.
   by unfold prot_plus in *; simpls; desf. 

  by eapply conform_upd_irr, conform_ext_irr; eauto; rewrite LEQ; ins.

Case "Aload ATna".

  exists (upd L (TsbS b) r); repeat (split; eauto using in_eq, upds, compat_guar1); ins;
    rupd; desc; try congruence.
  destruct IN; clarify; [|by apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr; rupd; vauto; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 

  by eapply conform_upd_irr, conform_ext_irr; eauto; rewrite LEQ; ins.

Case "Aload ATat".

  exists (upd L (TsbS b) r); repeat (split; eauto using in_eq, upds, compat_guar1); ins;
    rupd; desc; try congruence.
  destruct IN; clarify; [|by apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr; rupd; vauto; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 

  by eapply conform_upd_irr, conform_ext_irr; eauto; rewrite LEQ; ins.

Case "Astore ATna".

  exists (upd L (TsbS b) r_sb); repeat (split; eauto using in_eq, upds, compat_guar1); ins;
    rupd; desc; try congruence.
  destruct IN; clarify; [|by apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr; rupd; try done; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 

  eapply compat_guar1 in CC; subst.
  by eapply compat_guar2; eauto.

  apply conform_upd_irr; red; ins.
  destruct INa as [|INa]; clarify; [by rewrite LEQ in *|].
  destruct INb; clarify; eauto.
  assert (l = loc); subst.
    red in CONS; desc; clear G1; eapply (proj1 Cmo) in MO; desc.
    by rewrite LEQ in *; ins; desf; destruct (amap a); ins; desf.
  edestruct (CC (TrfS a :: TescS b0 :: nil)).
    by repeat constructor; ins; intro; desf.
   by clear; intros ? _; ins; desf.
  simpl; rupd.

  eapply V in INa; red in INa; desc.
  rewrite !res_plusA, !(res_plusAC r), <- (res_plus_strip_l (L (TrfS a))). 
  rewrite res_plusA, <- res_plusA, VrfS'.
  replace (r +!+ res_strip (out_rf (b0 :: acts) L a)) with Rundef; ins.
  revert G1; clear - GUAR ATa.
  destruct (amap a); simpls; desf; desf;
  red in AG; desc; clarify;
  clear - Grf; destruct r, (out_rf (b0 :: acts) L a); simpl; desf; ins;
  inv_lplus loc; rewrite prot_plusC, Grf in *; ins; desf; ins. 

Case "Astore ATat".

  exists (upd (upd L (TrfS b) r_rf) (TsbS b) r_sb),; ins;
    rupd; desc; try congruence.

  destruct IN; clarify; [|by repeat apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr in *; rupd; try done; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 

  eby edestruct NRF.

  assert (CO': conform (b :: acts) amap mo (upd L (TrfS b) r_rf) (b :: acts)).
  {
    eapply conform_suff_immediate; eauto.
      by rewrite LEQ; ins.
      by intro; eapply vghost_get_bot in VG; eauto; red in AG; desf; desf; congruence.
      by red in AG; desf; congruence.
      2: by eapply guar_sidecond; eauto.
    red in WPE; rewrite LEQ in WPE.
    ins; red in AG; desf.
      exploit visible_uninit; eauto using vghost_get_back_uninit.
      rewrite mhbE; intro; desf.
      red in CONS; desc; red in Cmo; desc.
      exfalso; red in IMM; desc.
      by eapply Cmo2, Cmo0, Cmo3; eauto; instantiate; rewrite LEQ.
    exploit WPE; clear WPE; eauto using vghost_get_back_at.
    intro WPE; desf.
    assert (a0 = a); desf; eauto.
    destruct (eqP a0 a); desf.
    by exfalso; red in CONS; desc; eapply Cmo in n; unfold immediate in *; desf; eauto.
  }

  repeat (split; eauto using in_eq, upds, compat_guar1, conform_upd_irr); ins;
    rupd; desc; try congruence.
  rewrite updC; ins; eapply compat_upd2_acc with (r := r_rf); eauto; ins.
  eapply compat_guar1 in CC.
  2: by replace (L (TrfS b)) with res_emp; eauto using res_plus_emp_l.
  assert (AG':=AG); red in AG; desc; rewrite REQ; desf. 
    by eapply compat_guar2; eauto.
  by eapply compat_atom_write; eauto; rewrite LEQ; ins; desf.

Case "Armw".

  exists (upd (upd L (TrfS b) r_rf) (TsbS b) r_sb). 

  assert (CO': conform (b :: acts) amap mo (upd L (TrfS b) r_rf) (b :: acts)).
  {
    eapply conform_suff_immediate; eauto.
      by rewrite LEQ; ins.
      by intro; eapply vghost_get_bot in VG; eauto; red in AG; desf; desf; congruence.
      by red in AG; desf; congruence.
      2: by eapply guar_sidecond; eauto.

    ins; red in AG; desf.
      exploit visible_uninit; eauto using vghost_get_back_uninit.
      rewrite mhbE; intro; desf.
      red in CONS; desc; red in Cmo; desc.
      exfalso; red in IMM; desc.
      by eapply Cmo2, Cmo0, Cmo3; eauto; instantiate; rewrite LEQ.

    eapply prot_trans_ok_trans with (res_get r l).
    2: rewrite GET, Grf; split; ins.
    2: by eexists; rewrite SEQ; eauto.
  
    assert (RF: rf b = Some a).
      edestruct visible_atomic with (l:=l); try solve [clear CC; eauto]; 
        eauto using vghost_get_back_at; desf.
      clear - H H0 CONS LEQ IMM AW.
      red in CONS; desc; specialize (Crf b); desf; ins; desf.
      2: by edestruct Crf; eauto; rewrite LEQ; ins.
      rewrite LEQ in *; ins; desf.
      exploit Crmw; eauto; instantiate; try rewrite LEQ; try done; intros IMM'. 
      destruct (eqP a0 a) as [|NEQ]; desf.
      by eapply Cmo in NEQ; eauto; desf; exfalso; unfold immediate in *; desc; eauto.
  
    specialize (V a IN); red in V; desc.
    assert (OG: exists tau S, res_get (out_rf (b :: acts) L a) l = p_at tau S).
      by clear - GUAR AW; destruct (amap a); ins; desf; desf; try red in AG; desf; eauto.
    desf; revert OG.
    rewrite <- VrfM with (c:=b); eauto using in_eq; try (by rewrite LEQ).
    intro; rewrite OG, GET.
    edestruct res_lt_rf_in with (L:=L) (a:=a) (b:=b) (acts:=b::acts) as (rrr & XX); 
      eauto using in_cons.
    rewrite <- XX in *.
    edestruct (res_get_plus _ _ rrr OG) as (?&?&?).
      by intro A; rewrite A in *; red in VG; desf.
    eapply vghost_get_at in VG; eauto.
    rewrite VG in *; desf; split; ins; red; eauto using prot_trans_refl.
  }

  repeat (split; eauto using in_eq, upds, compat_guar1, conform_upd_irr); ins;
    rupd; desc; try congruence.
  destruct IN; clarify; [|by repeat apply valid_node_upd_irr; eauto; intro; desf].

  red; unfold NW; eexists r, I; rewrite LEQ; rupd; repeat (split; ins; rupd; try edone);
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *; 
  rewrite ?map_upd_irr, ?escr_ffun_upd_irr in *; rupd; vauto; 
    try (by rewrite in_map_iff; intro; desf);
    try (by ins; rewrite ?Osb0, ?Orf, ?Orf0, ?res_plus_emp_r; try done; apply LR). 
  eby edestruct NRF. 

  rewrite updC; ins.
  eapply compat_upd2_acc with (r := r_rf); eauto; ins.
  eapply compat_guar1 in CC.
  2: by replace (L (TrfS b)) with res_emp; eauto using res_plus_emp_l.
  assert (AG':=AG); red in AG; desc; rewrite REQ; desf. 
    by eapply compat_guar2; eauto.
  by eapply compat_atom_write; eauto; rewrite LEQ; ins; desf.
Qed.

(******************************************************************************)
(** * 3.3.6. Prepare *)
(******************************************************************************)

Lemma prepare_step :
  forall b acts amap sb rf mo
    (CONS: Consistent (b :: acts) amap sb rf mo) 
    (NIN : ~ In b acts)
    (ND  : NoDup (b :: acts)) a
    (SB  : sb a b) amap_old sb_old rf_old mo_old
    (OCON: Consistent acts amap_old sb_old rf_old mo_old) L 
    (LR  : label_restr acts amap_old sb_old rf_old L) r r_rem
    (LAB : L (TsbS a) = r +!+ r_rem)
    (V   : forall a, In a acts -> valid_node acts acts amap_old rf_old L a)
    (CC  : compat amap_old sb_old rf_old L)
    (CO  : conform acts amap_old mo_old L acts)
    (Ea  : forall x, In x acts -> amap x = amap_old x)
    (Esb : forall x y, sb x y <-> sb_old x y \/ x = a /\ y = b)
    (Erf : forall x, x <> b -> rf x = rf_old x)
    (Emo : forall x, x <> b -> forall y, y <> b -> (mo x y <-> mo_old x y)),
  exists L', 
    << V   : forall a, In a acts -> valid_node (b :: acts) acts amap rf L' a >> /\
    << CC  : compat amap sb rf L' >> /\
    << CO  : conform (b :: acts) amap mo L' acts >> /\
    << LAB : L' (TsbS a) = r_rem >> /\
    << LSB : in_sb (b :: acts) L' b = r >> /\
    << LR  : label_restr (b :: acts) amap sb rf L' >> /\
    << Irf : in_rf (b :: acts) L' b = res_emp >> /\
    << Iesc: in_esc (b :: acts) L' b = res_emp >> /\
    << Isb : forall a', a' <> a -> L' (Tsb a' b) = res_emp >> /\
    << O   : outgoing (b :: acts) L' b = res_emp >> /\
    << LEQ : forall a', a' <> a -> L' (TsbS a') = L (TsbS a') >> /\
    << PC  : prefix_closed sb rf acts >>.
Proof.
  ins; unfold NW.
  assert (NEQ: a <> b).
    by intro; subst; red in CONS; desc; edestruct ACYC; vauto.
  exists (upd (upd L (Tsb a b) r) (TsbS a) r_rem).

  assert (INa : In a acts).
    by red in CONS; desc; apply FIN, proj1 in SB; ins; desf.

  assert (SBbx : forall x, L (Tsb b x) = res_emp).
    by intros; apply LR; intro X; eapply (proj2 (proj1 OCON)) in X; desf.
  assert (SBxb : forall x, L (Tsb x b) = res_emp).
    by intros; apply LR; intro X; eapply (proj2 (proj1 OCON)) in X; desf.

  assert (RFbx : forall x, L (Trf b x) = res_emp).
    by intros; apply LR; intro X; eapply (Consistent_sw_in OCON) in X; desf. 
  assert (RFxb : forall x, L (Trf x b) = res_emp).
    by intros; apply LR; intro X; eapply (Consistent_sw_in OCON) in X; desf. 

  assert (ESCbx : forall x, L (Tesc b x) = res_emp).
    by intros; apply LR; intro X; eapply (Consistent_hb_in OCON) in X; desf. 
  assert (ESCxb : forall x, L (Tesc x b) = res_emp).
    by intros; apply LR; intro X; eapply (Consistent_hb_in OCON) in X; desf. 

  assert (SBimp: forall a b, sb_old a b -> sb a b).
    by intros; rewrite Esb; vauto.

  assert (SWimp: forall a b,
       synchronizes_with amap_old rf_old a b ->
       synchronizes_with amap rf a b).
    ins; red; red in H; desf. 
    assert (In a0 acts) by (by apply (proj1 (proj1 OCON)); intro X; rewrite X in *).
    assert (In b0 acts) by (by apply (proj1 (proj1 OCON)); intro X; rewrite X in *).
    rewrite Erf, !Ea; ins; congruence.

  assert (HBimp: forall a b,
       happens_before amap_old sb_old rf_old a b ->
       happens_before amap sb rf a b).
    by induction 1; unfold imm_happens_before in *; desf; 
       eauto using hb_trans, hb_sw, hb_sb. 

  split; ins.
  red; unfold NW.
  specialize (V _ H); red in V; desf; exists r0, I.
  rewrite replace_out_sb; vauto.
  unfold incoming, out_rf, out_esc, in_rf, in_esc in *.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; ins; desf; congruence). 
  unfold in_sb, out_sb in *; ins; rupd; try congruence.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; ins; desf; congruence). 
  rewrite !SBbx, !RFbx, !ESCbx, SBxb, RFxb, ESCxb, !res_plus_emp_l; ins.
  rewrite (Ea _ H), !escr_ffun_upd_irr; try done.
  repeat (split; try done).
    ins; rupd; rewrite Ea in RMW; rewrite Erf in RF; ins; eauto; congruence. 
    ins; rupd; eapply VrfS; ins; rewrite <- Ea in RMW; rewrite <- Erf in RF; eauto; congruence.

  eapply res_lt_trans; eauto; generalize I as J; induction J; 
    ins; desf; ins; eauto using res_lt_refl, res_lt_frame_l, res_lt_plus_l. 
  by exfalso; unfold escr_ffun in *; desf; desf; destruct n; 
     repeat eexists; ins; eauto. 
  by rewrite LAB, SBxb, res_plus_emp_l.

  split. 
    rewrite updC; ins; apply compat_upd2; ins.
        red; ins; eapply CC; eauto; ins; eapply IND with (te:=te) (te':=te'); eauto.
        by rewrite mhbE in *; desf; eauto.
      by rewrite LAB, res_plusC; vauto. 
    by rewrite LAB, SBxb, (res_plusC r); vauto. 

  split; rupd.
    red; ins.
    rewrite Emo in *; try congruence.
    rewrite (Ea _ INa0), (Ea _ INb) in *.
    exploit CO; try exact MO; eauto.
    by unfold out_rf; ins; rupd; rewrite !RFxb, !map_upd_irr, !res_plus_emp_l;
       ins; rewrite in_map_iff; intro; desf.

  unfold outgoing, out_sb, out_rf, out_esc, in_rf, in_esc; ins; rupd; try congruence.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; ins; desf; congruence). 
  rewrite !SBbx, !RFbx, !ESCbx, !res_plus_emp_l.
  assert (LR': label_restr (b :: acts) amap sb rf
             (upd (upd L (Tsb a b) r) (TsbS a) r_rem)).
    apply lr_upd_sbS; auto using in_cons.
    apply lr_upd_sb; auto using in_cons.
    by red; unfold NW; repeat split; ins; apply LR; 
       intro; desf; destruct H; auto.

  repeat (split; try done); unfold NW; ins; rupd; eauto using rsum_map_map_eq_empI; try congruence.

  unfold in_sb; ins; rupd; try congruence; rewrite SBxb, res_plus_emp_l.
  inv ND; eapply In_NoDup_Permutation in INa; eauto; desf.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ INa))); ins; rupd.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; ins; desf; congruence). 
  by rewrite rsum_map_map_eq_empI, res_plus_emp_r; ins.

  by repeat first [apply rsum_map_map_eq_empI; ins | apply res_plus_eq_empI| by apply LR].

  rewrite Esb in SB0; desf.
  by eapply (Consistent_prefix_closed OCON) in SB0; eauto.

  rewrite Erf in RF; desf; try congruence.
  by eapply (Consistent_prefix_closed OCON) in RF; eauto.
Qed.
